@@ -19,17 +19,17 @@ Core Pylightnix definitions
 from pylightnix.imports import ( sha256, deepcopy, isdir, islink, makedirs,
     join, json_dump, json_load, json_dumps, json_loads, isfile, relpath,
     listdir, rmtree, mkdtemp, replace, environ, split, re_match, ENOTEMPTY,
-    get_ident, contextmanager, OrderedDict, lstat, maxsize, readlink )
+    get_ident, contextmanager, OrderedDict, lstat, maxsize, readlink, chain )
 from pylightnix.utils import ( dirhash, assert_serializable, assert_valid_dict,
     dicthash, scanref_dict, scanref_list, forcelink, timestring, parsetime,
     datahash, readjson, tryread, encode, dirchmod, dirrm, filero, isrref,
-    isdref, traverse_dict, ispromise, isclaim )
-from pylightnix.types import (
-    Dict, List, Any, Tuple, Union, Optional, Iterable, IO, Path, Hash, DRef,
-    RRef, RefPath, PromisePath, HashPart, Callable, Context, Name, NamedTuple,
-    Build, RConfig, ConfigAttrs, Derivation, Stage, Manager, Matcher, Realizer,
-    Set, Closure, Generator, Key, TypeVar, BuildArgs, PYLIGHTNIX_PROMISE_TAG,
-    PYLIGHTNIX_CLAIM_TAG, Config, RealizeArg, InstantiateArg )
+    isdref, traverse_dict, ispromise, isclaim, tryread_def, tryreadjson_def )
+from pylightnix.types import ( Dict, List, Any, Tuple, Union, Optional,
+    Iterable, IO, Path, Hash, DRef, RRef, RefPath, PromisePath, HashPart,
+    Callable, Context, Name, NamedTuple, Build, RConfig, ConfigAttrs,
+    Derivation, Stage, Manager, Matcher, Realizer, Set, Closure, Generator,
+    Key, BuildArgs, PYLIGHTNIX_PROMISE_TAG, PYLIGHTNIX_CLAIM_TAG, Config,
+    RealizeArg, InstantiateArg, Tag, Group, RRefGroup )
 
 #: *Do not change!*
 #: Tracks the version of pylightnix storage
@@ -56,6 +56,10 @@ PYLIGHTNIX_STORE = join(PYLIGHTNIX_ROOT, f'store-v{PYLIGHTNIX_STORE_VERSION}')
 #: Set the regular expression pattern for valid name characters.
 PYLIGHTNIX_NAMEPAT = "[a-zA-Z0-9_-]"
 
+#: Reserved file names which are treated specially be the core. Users should
+#: not normally create or alter files with this names.
+PYLIGHTNIX_RESERVED = ['context.json','group.json']
+
 #  ____       __
 # |  _ \ ___ / _|___
 # | |_) / _ \ |_/ __|
@@ -92,14 +96,32 @@ def mkname(s:str)->Name:
   assert_valid_name(Name(s))
   return Name(s)
 
-def path2rref(p:Path)->RRef:
+def path2rref(p:Path)->Optional[RRef]:
+  """ Takes either a system path of some realization in the Pylightnix storage
+  or a symlink pointing to such path. Return `RRef` which corresponds to this
+  path.
+
+  Note: `path2rref` doesn't actually check the existance of such an object in
+  storage """
   if islink(p):
     p=Path(readlink(p))
   head,h1=split(p)
-  _,dref=split(head)
-  h2,nm=undref(DRef('dref:'+dref))
+  _,dref_part=split(head)
+  dref=DRef('dref:'+dref_part)
+  if not isdref(dref):
+    return None
+  h2,nm=undref(dref)
   return mkrref(HashPart(h1),HashPart(h2),mkname(nm))
 
+def mktag(s:str)->Tag:
+  for c in ['\n',' ']:
+    assert c not in s
+  return Tag(s)
+
+def mkgroup(s:str)->Group:
+  for c in ['\n',' ']:
+    assert c not in s
+  return Group(s)
 
 #   ____             __ _
 #  / ___|___  _ __  / _(_) __ _
@@ -288,34 +310,45 @@ def store_drefs()->Iterable[DRef]:
     if dirname[-4:]!='.tmp' and isdir(join(PYLIGHTNIX_STORE,dirname)):
       yield mkdref(HashPart(dirname[:32]), Name(dirname[32+1:]))
 
-def store_rrefs_(dref:DRef)->Iterable[RRef]:
+def rrefs2groups(rrefs:List[RRef])->List[RRefGroup]:
+  return [({store_tag(rref):rref for rref in rrefs if store_group(rref)==g}) \
+      for g in sorted({store_group(rref) for rref in rrefs})]
+
+def groups2rrefs(grs:List[RRefGroup])->List[RRef]:
+  return list(chain.from_iterable([gr.values() for gr in grs]))
+
+def store_rrefs_(dref:DRef)->List[RRefGroup]:
   """ Iterate over all realizations of a derivation `dref`. The sort order is
   unspecified. """
   (dhash,nm)=undref(dref)
   drefpath=store_dref2path(dref)
+  rrefs:List[RRef]=[]
   for f in listdir(drefpath):
     if f[-4:]!='.tmp' and isdir(join(drefpath,f)):
-      yield mkrref(HashPart(f), dhash, nm)
+      rrefs.append(mkrref(HashPart(f), dhash, nm))
+  return rrefs2groups(rrefs)
 
-def store_rrefs(dref:DRef, context:Context)->Iterable[RRef]:
-  """ Iterate over realizations of a derivation `dref`, which match a
-  [context](#pylightnix.types.Context). The sort order is unspecified. """
-  for rref in store_rrefs_(dref):
-    context2=store_context(rref)
+def store_rrefs(dref:DRef, context:Context)->List[RRefGroup]:
+  """ Iterate over realizations of a derivation `dref` which match a specified
+  [context](#pylightnix.types.Context). Sorting order is unspecified. """
+  rgs:List[RRefGroup]=[]
+  for rg in store_rrefs_(dref):
+    context2=store_context(list(rg.values())[0])
     if context_eq(context,context2):
-      yield rref
+      rgs.append(rg)
+  return rgs
 
-def store_deref_(context_holder:RRef, dref:DRef)->List[RRef]:
+def store_deref_(context_holder:RRef, dref:DRef)->List[RRefGroup]:
   return context_deref(store_context(context_holder), dref)
 
-def store_deref(context_holder:RRef, dref:DRef)->RRef:
+def store_deref(context_holder:RRef, dref:DRef)->RRefGroup:
   """ For any realization `context_holder` and it's dependency `dref`, `store_deref`
   queries the realization reference of this dependency.
 
   See also [build_deref](#pylightnix.core.build_deref)"""
-  rrefs=store_deref_(context_holder, dref)
-  assert len(rrefs)==1
-  return rrefs[0]
+  rgs=store_deref_(context_holder, dref)
+  assert len(rgs)==1
+  return rgs[0]
 
 def store_buildtime(rref:RRef)->Optional[str]:
   """ Return the buildtime of the current RRef in a format specified by the
@@ -328,6 +361,15 @@ def store_buildtime(rref:RRef)->Optional[str]:
   realizations may not provide this information. """
   return tryread(Path(join(rref2path(rref),'__buildtime__.txt')))
 
+def store_tag(rref:RRef)->Tag:
+  """ Realizations may be marked with a tag. By default the tag is set to be
+  'out'. """
+  return mktag(tryreadjson_def(Path(join(rref2path(rref),'group.json')),{}).get('tag','out'))
+
+def store_group(rref:RRef)->Group:
+  """ Return group identifier of the realization """
+  return mkgroup(tryreadjson_def(Path(join(rref2path(rref),'group.json')),{}).get('group',rref))
+
 def store_gc(keep_drefs:List[DRef], keep_rrefs:List[RRef])->Tuple[Set[DRef],Set[RRef]]:
   """ Take roots which are in use and should not be removed. Return roots which
   are not used and may be removed. Actual removing is to be done by the user.
@@ -343,9 +385,10 @@ def store_gc(keep_drefs:List[DRef], keep_rrefs:List[RRef])->Tuple[Set[DRef],Set[
   for dref in store_drefs():
     if dref not in closure_drefs:
       remove_drefs.add(dref)
-    for rref in store_rrefs_(dref):
-      if rref not in closure_rrefs:
-        remove_rrefs.add(rref)
+    for rg in store_rrefs_(dref):
+      for rref in rg.values():
+        if rref not in closure_rrefs:
+          remove_rrefs.add(rref)
   return remove_drefs,remove_rrefs
 
 
@@ -384,7 +427,7 @@ def store_instantiate(c:Config)->DRef:
       raise
   return dref
 
-def store_realize(dref:DRef, l:Context, o:Path)->RRef:
+def store_realize_tag(dref:DRef, l:Context, o:Path, leader:Optional[Tuple[Tag,RRef]]=None)->RRef:
   """
   FIXME: Assert or handle possible but improbable hash collision (*)
   """
@@ -395,12 +438,21 @@ def store_realize(dref:DRef, l:Context, o:Path)->RRef:
   assert isdir(o), (
      f"While realizing {dref}: Outpath is expected to be a path to existing "
      f"directory, but got {o}")
-  assert not isfile(join(o,'context.json')), (
-     f"While realizing {dref}: one of build artifacts has name 'context.json'. "
-     f"This name is reserved, please rename the artifact.")
+
+  for fn in PYLIGHTNIX_RESERVED:
+    assert not isfile(join(o,fn)), (
+       f"While realizing {dref}: output folder '{o}' contains file '{fn}'. "
+       f"This name is reserved, please use another name. List of reserved "
+       f"names: {PYLIGHTNIX_RESERVED}")
+
   with open(join(o,'context.json'), 'w') as f:
     f.write(context_serialize(l))
 
+  if leader is not None:
+    tag,group_rref=leader
+    with open(join(o,'group.json'), 'w') as f:
+      json_dump({'tag':tag,'group':group_rref},f)
+
   rhash=dirhash(o)
   rref=mkrref(trimhash(rhash),dhash,nm)
   rrefpath=rref2path(rref)
@@ -422,152 +474,13 @@ def store_realize(dref:DRef, l:Context, o:Path)->RRef:
       raise
   return rref
 
-#  ____        _ _     _
-# | __ ) _   _(_) | __| |
-# |  _ \| | | | | |/ _` |
-# | |_) | |_| | | | (_| |
-# |____/ \__,_|_|_|\__,_|
-
-def mkbuildargs(dref:DRef, context:Context, timeprefix:Optional[str],
-    iarg:InstantiateArg, rarg:RealizeArg)->BuildArgs:
-  assert_valid_config(store_config(dref))
-  return BuildArgs(dref, context, timeprefix, iarg, rarg)
-
-def mkbuild(dref:DRef, context:Context, buildtime:bool=True)->Build:
-  timeprefix=timestring() if buildtime else None
-  return Build(mkbuildargs(dref,context,timeprefix,{},{}))
-
-B=TypeVar('B')
-
-def build_wrapper_(
-    f:Callable[[B],None],
-    ctr:Callable[[BuildArgs],B],
-    buildtime:bool=True)->Realizer:
-  def _wrapper(dref,context,rarg)->List[Path]:
-    timeprefix=timestring() if buildtime else None
-    b=ctr(mkbuildargs(dref,context,timeprefix,{},rarg));
-    try:
-      f(b);
-      return list(getattr(b,'outpaths'))
-    except:
-      print(f'Build wrapper of {dref} raised an exception. Remaining '
-            f'build directories are: {getattr(b,"outpaths","<unknown>")}')
-      raise
-  return _wrapper
-
-def build_wrapper(f:Callable[[Build],None], buildtime:bool=True)->Realizer:
-  """ Build Adapter which convers user-defined realizers which use
-  [Build](#pylightnix.types.Build) API into a low-level
-  [Realizer](#pylightnix.types.Realizer) """
-  return build_wrapper_(f,Build,buildtime)
-
-def build_config(b:Build)->RConfig:
-  """ Return the [RConfig](#pylightnix.types.RConfig) object of the realization
-  being built. """
-  return store_config(b.dref)
-
-def build_context(b:Build)->Context:
-  """ Return the [Context](#pylightnix.types.Context) object of the realization
-  being built. """
-  return b.context
-
-def build_cattrs(b:Build)->Any:
-  """ Cache and return `ConfigAttrs`. Cache allows realizers to update it's
-  value during the build process, e.g. to use it as a storage. """
-  if b.cattrs_cache is None:
-    b.cattrs_cache=config_cattrs(build_config(b))
-  return b.cattrs_cache
-
-def build_setoutpaths(b:Build, nouts:int)->List[Path]:
-  assert nouts>0, f"Attempt to set {nouts} (<=0) output paths"
-  assert len(b.outpaths)==0, f"Build outpaths were already set to {b.outpaths}"
-  prefix=f'{b.timeprefix}_{config_hash(build_config(b))[:8]}_'
-  outpaths=[Path(mkdtemp(prefix=prefix, dir=PYLIGHTNIX_TMP)) for _ in range(nouts)]
-  if b.timeprefix is not None:
-    for outpath in outpaths:
-      with open(join(outpath,'__buildtime__.txt'), 'w') as f:
-        f.write(b.timeprefix)
-  b.outpaths=outpaths
-  return b.outpaths
-
-def build_outpaths(b:Build)->List[Path]:
-  assert len(b.outpaths)>0, f"Build outpaths were not set"
-  return b.outpaths
-
-def build_outpath(b:Build)->Path:
-  """ Return the output path of the realization being built. Output path is a
-  path to valid temporary folder where user may put various build artifacts.
-  Later this folder becomes a realization. """
-  if len(b.outpaths)==0:
-    return build_setoutpaths(b,1)[0]
-  paths=build_outpaths(b)
-  assert len(paths)==1, f"Build was set to have multiple output paths: {paths}"
-  return paths[0]
-
-def build_name(b:Build)->Name:
-  """ Return the name of a derivation being built. """
-  return Name(config_name(build_config(b)))
-
-def build_deref_(b:Build, dref:DRef)->List[RRef]:
-  """ For any [realization](#pylightnix.core.realize) process described with
-  it's [Build](#pylightnix.types.Build) handler, `build_deref` queries a
-  realization of dependency `dref`.
-
-  `build_deref` is designed to be called from
-  [Realizer](#pylightnix.types.Realizer) functions. In other cases,
-  [store_deref](#pylightnix.core.store_deref) should be used.  """
-  return context_deref(build_context(b), dref)
-
-def build_deref(b:Build, dref:DRef)->RRef:
-  rrefs=build_deref_(b,dref)
-  assert len(rrefs)==1
-  return rrefs[0]
-
-def build_paths(b:Build, refpath:RefPath)->List[Path]:
-  """ Convert given [RefPath](#pylightnix.types.RefPath) (which may be either a
-  regular RefPath or an ex-[PromisePath](#pylightnix.types.PromisePath)) into
-  one or many filesystem paths. Conversion refers to the
-  [Context](#pylightnix.types.Context) of the current realization process by
-  accessing it's [build_context](#pylightnix.core.build_context).
-
-  Typically, we configure stages to match only one realization at once, so the
-  returned list often has only one entry. Consider using
-  [build_path](#pylightnix.core.build_path) if this fact is known in advance.
-
-  Example:
-  ```python
-  def config(dep:DRef)->RConfig:
-    name = 'example-stage'
-    input = [dep,"path","to","input.txt"]
-    output = [promise,"output.txt"]
-    some_param = 42
-    return mkconfig(locals())
-
-  def realize(b:Build)->None:
-    c=config_cattrs(b)
-    with open(build_path(b, c.input),'r') as finp:
-      with open(build_path(b, c.output),'w') as fout:
-        fout.write(finp.read())
-
-  def mystage(m:Manager)->DRef:
-    dep:DRef=otherstage(m)
-    return mkdrv(m, config(dep), match_only(), build_wrapper(realize))
-  ```
-  """
-  assert_valid_refpath(refpath)
-  if refpath[0]==b.dref:
-    assert len(b.outpaths), (
-        f"Attempt to access build outpaths before they are set. Call"
-        f"`build_outpath(b,num)` first to set their number." )
-    return [Path(join(path, *refpath[1:])) for path in b.outpaths]
-  else:
-    return [Path(join(rref2path(rref), *refpath[1:])) for rref in build_deref_(b, refpath[0])]
-
-def build_path(b:Build, refpath:RefPath)->Path:
-  """ A single-realization version of the [build_paths](#pylightnix.core.build_paths). """
-  paths=build_paths(b,refpath)
-  assert len(paths)==1
-  return paths[0]
+def store_realize_group(dref:DRef, l:Context, og:Dict[Tag,Path])->RRefGroup:
+  rrefg={}
+  rrefg[Tag('out')]=store_realize_tag(dref, l, og[Tag('out')])
+  for tag,o in og.items():
+    if tag!=Tag('out'):
+      rrefg[tag]=store_realize_tag(dref,l,o,(tag,rrefg[Tag('out')]))
+  return rrefg
 
 #   ____            _            _
 #  / ___|___  _ __ | |_ _____  _| |_
@@ -592,11 +505,11 @@ def context_add(context:Context, dref:DRef, rrefs:List[RRef])->Context:
   context[dref]=rrefs
   return context
 
-def context_deref(context:Context, dref:DRef)->List[RRef]:
+def context_deref(context:Context, dref:DRef)->List[RRefGroup]:
   assert dref in context, (
       f"Context {context} doesn't declare {dref} among it's dependencies so we "
       f"can't dereference it." )
-  return context[dref]
+  return rrefs2groups(context[dref])
 
 def context_serialize(c:Context)->str:
   assert_valid_context(c)
@@ -649,6 +562,7 @@ def mkdrv(m:Manager,
   derivation produced.
 
   Arguments:
+  - `m:Manager`: A Manager to update with a new derivation
   - `check_promises:bool=True`: Make sure that all
     [PromisePath](#pylightnix.types.PromisePaths) of stage's configuration
     correspond to existing files or firectories.
@@ -670,12 +584,12 @@ def mkdrv(m:Manager,
            f"RConfig:\n{store_config_(dref)}"))
 
   def _promise_aware(realizer)->Realizer:
-    def _realizer(dref:DRef,ctx:Context,rarg:RealizeArg)->List[Path]:
-      outpaths=realizer(dref,ctx,rarg)
+    def _realizer(dref:DRef,ctx:Context,rarg:RealizeArg)->List[Dict[Tag,Path]]:
+      outgroups=realizer(dref,ctx,rarg)
       for key,refpath in config_promises(store_config_(dref),dref):
-        for o in outpaths:
-          assert_promise_fulfilled(key,refpath,o)
-      return outpaths
+        for g in outgroups:
+          assert_promise_fulfilled(key,refpath,g[Tag('out')])
+      return outgroups
     return _realizer
 
   m.builders[dref]=Derivation(dref=dref,
@@ -734,7 +648,9 @@ def instantiate(stage:Any, *args, **kwargs)->Closure:
   """
   return instantiate_(Manager(), stage, *args, **kwargs)
 
-RealizeSeqGen = Generator[Tuple[DRef,Context,Derivation,RealizeArg],Tuple[Optional[List[RRef]],bool],List[RRef]]
+RealizeSeqGen = Generator[Tuple[DRef,Context,Derivation,RealizeArg],
+                          Tuple[Optional[List[RRefGroup]],bool],
+                          List[RRefGroup]]
 
 def realize(closure:Closure, force_rebuild:Union[List[DRef],bool]=[],
                              assert_realized:List[DRef]=[])->RRef:
@@ -793,6 +709,7 @@ def realizeMany(closure:Closure, force_rebuild:Union[List[DRef],bool]=[],
   - FIXME: Update derivation's matcher after forced rebuilds. Matchers should
     remember and reproduce user's preferences.
   """
+  res:List[RRefGroup]
   force_interrupt:List[DRef]=[]
   if isinstance(force_rebuild,bool):
     if force_rebuild:
@@ -810,7 +727,7 @@ def realizeMany(closure:Closure, force_rebuild:Union[List[DRef],bool]=[],
       gen.send((None,False)) # Ask for default action
   except StopIteration as e:
     res=e.value
-  return res
+  return groups2rrefs(res)
 
 def realizeSeq(closure:Closure, force_interrupt:List[DRef]=[],
                                 assert_realized:List[DRef]=[],
@@ -830,46 +747,47 @@ def realizeSeq(closure:Closure, force_interrupt:List[DRef]=[],
     target_deps=store_deepdeps([target_dref])
     for drv in closure.derivations:
       dref=drv.dref
+      rrefgs:Optional[List[RRefGroup]]
       if dref in target_deps or dref==target_dref:
         c=store_config(dref)
         dref_deps=store_deepdeps([dref])
         dref_context={k:v for k,v in context_acc.items() if k in dref_deps}
         if dref in force_interrupt_:
-          rrefs,abort=yield (dref,dref_context,drv,realize_args.get(dref,{}))
+          rrefgs,abort=yield (dref,dref_context,drv,realize_args.get(dref,{}))
           if abort:
             return []
         else:
-          rrefs=drv.matcher(dref,dref_context)
-        if rrefs is None:
+          rrefgs=drv.matcher(dref,dref_context)
+        if rrefgs is None:
           assert dref not in assert_realized, (
             f"Stage '{dref}' was assumed to be already realized. "
             f"Unfortunately, it is not the case. Config:\n"
             f"{store_config(dref)}"
             )
-          paths=drv.realizer(dref,dref_context,realize_args.get(dref,{}))
-          rrefs_built=[store_realize(dref,dref_context,path) for path in paths]
-          rrefs_matched=drv.matcher(dref,dref_context)
-          assert rrefs_matched is not None, (
+          gpaths:List[Dict[Tag,Path]]=drv.realizer(dref,dref_context,realize_args.get(dref,{}))
+          rrefgs_built=[store_realize_group(dref,dref_context,g) for g in gpaths]
+          rrefgs_matched=drv.matcher(dref,dref_context)
+          assert rrefgs_matched is not None, (
             f"Derivation {dref}: Matcher repeatedly asked the core to "
             f"realize. Probably, realizer doesn't match the matcher. "
             f"In particular, the follwoing just-built rrefs are "
-            f"unmatched: {rrefs_built}" )
-          if (set(rrefs_built) & set(rrefs_matched)) == set() and \
-             (set(rrefs_built) | set(rrefs_matched)) != set():
+            f"unmatched: {rrefgs_built}" )
+          if (set(groups2rrefs(rrefgs_built)) & set(groups2rrefs(rrefgs_matched))) == set() and \
+             (set(groups2rrefs(rrefgs_built)) | set(groups2rrefs(rrefgs_matched))) != set():
             print(f"Warning: None of the newly obtained realizations of "
                   f"{dref} were matched by the matcher. To capture those "
                   f"realizations explicitly, try `matcher([exact(..)])`")
-          rrefs=rrefs_matched
-        context_acc=context_add(context_acc,dref,rrefs)
-    assert rrefs is not None
-    return rrefs
+          rrefgs=rrefgs_matched
+        context_acc=context_add(context_acc,dref,groups2rrefs(rrefgs))
+    assert rrefgs is not None
+    return rrefgs
 
 
 def mksymlink(rref:RRef, tgtpath:Path, name:str, withtime=True)->Path:
   """ Create a symlink pointing to realization `rref`. Other arguments define
   symlink name and location. Informally,
   `{tgtpath}/{timeprefix}{name} --> $PYLIGHTNIX_STORE/{dref}/{rref}`.
-  Overwrite existing symlinks.
+  Overwrite existing symlinks. Folder named `tgtpath` should exist.
   """
   assert_valid_rref(rref)
   assert isdir(tgtpath), f"Target link directory doesn't exist: '{tgtpath}'"
@@ -894,6 +812,11 @@ def match(keys:List[Key],
   """ Create a [Matcher](#pylightnix.types.Matcher) by combining different
   sorting keys and selecting a top-n threshold.
 
+  Only realizations which have [tag](#pylightnix.types.Tag) 'out' (which is a
+  default tag name) participate in matching. After the matching, Pylightnix
+  adds all non-'out' realizations which share [group](#pylightnix.types.Group)
+  with at least one matched realization.
+
   Arguments:
   - `keys`: List of [Key](#pylightnix.types.Key) functions. Defaults ot
   - `rmin`: An integer selecting the minimum number of realizations to accept.
@@ -909,13 +832,17 @@ def match(keys:List[Key],
   assert not ((rmax is None) and exclusive), (
     "Specifying `exclusive` has no effect without specifying `rmax`.")
   keys=keys+[texthash()]
-  def _matcher(dref:DRef, context:Context)->Optional[List[RRef]]:
-    keymap={}
-    rrefs=list(store_rrefs(dref, context))
-    for rref in rrefs:
-      keymap[rref]=[k(rref) for k in keys]
-    res=sorted(filter(lambda rref: None not in keymap[rref], rrefs),
-          key=lambda rref:keymap[rref], reverse=True)
+  def _matcher(dref:DRef, context:Context)->Optional[List[RRefGroup]]:
+    # Find 'out' RRefs in each group
+    grefs={gr[Tag('out')]:gr for gr in store_rrefs(dref,context)}
+
+    # Match only among realizations tagged as 'out'
+    keymap={rref:[k(rref) for k in keys] for rref in grefs.keys()}
+
+    # Apply filters and filter outputs
+    res:List[RRef]=sorted(filter(lambda rref: None not in keymap[rref], grefs.keys()),
+                          key=lambda rref:keymap[rref], reverse=True)
+    # Filter by range
     if rmin is not None:
       if not rmin<=len(res):
         return None
@@ -923,7 +850,9 @@ def match(keys:List[Key],
       if not len(res)<=rmax:
         assert not exclusive
         res=res[:rmax]
-    return res
+
+    # Return matched groups
+    return [grefs[rref] for rref in res]
   return _matcher
 
 
@@ -1053,21 +982,16 @@ def warn_rref_deps(c:Config)->None:
   if len(rrefs)>0:
     print(f"Warning: RRef dependencies were found in config {config_dict(c)}:\n{rrefs}")
 
-# def assert_no_rref_deps(c:RConfig)->None:
-#   _,rrefs=scanref_dict(config_dict(c))
-#   assert len(rrefs)==0, (
-#       f"RRef dependencies are forbidden, but config {config_dict(c)} containes {rrefs}")
-
 def assert_have_realizers(m:Manager, drefs:List[DRef])->None:
   have_drefs=set(m.builders.keys())
   need_drefs=store_deepdeps(drefs)|set(drefs)
   missing=list(need_drefs-have_drefs)
   assert len(missing)==0, (
-      f"Some derivations don't have realizers associated with them:\n"
+      f"The following derivations don't have realizers associated with them:\n"
       f"{missing}\n"
-      f"Did you take those DRefs from another manager?")
+      f"Did you pass those DRefs from another `instantiate` session ?")
 
-def assert_recursion_manager_empty():
+def assert_recursion_manager_empty()->None:
   global PYLIGHTNIX_RECURSION
   stack=PYLIGHTNIX_RECURSION.get(get_ident(),[])
   assert len(stack)==0
